# -*- coding: utf-8 -*-
# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)

"""
core codes of the mesh_illustris package.
"""

import os
import numpy as np
import h5py
import time

from .il_util import *
from .mesh import Mesh

__all__ = ["Dataset", "SingleDataset"]

class Dataset(object):
    """Dataset class stores a snapshot of simulation."""

    def __init__(self, datasets, n_chunk):
        """
        Args:
            datasets (list of SingleDataset): Chunks that store the 
                snapshot of simulation. 
            n_chunk (int): Number of chunks
        """

        super(Dataset, self).__init__()
        self._datasets = datasets
        self._n_chunk = n_chunk

    @property
    def datasets(self):
        """list of SingleDataset: Chunks that store the snapshot 
            of simulation."""
        return self._datasets

    @property
    def n_chunk(self):
        """int: Number of chunks."""
        return self._n_chunk

    def _combine(self, func, partType, fields, mdi=None, 
        float32=False, **kwargs):
        """
        Combine subsets (e.g., a box or sphere) of data in different chunks 
        into one subset.

        Args:
            func (str): Types of subset, must be "box" or "sphere". This 
                argument determines the slicing method.
            partType (str or list of str): Particle types to be loaded.
            fields (str or list of str): Particle fields to be loaded.
            mdi (None or list of int, default to None): sub-indeces to be 
                loaded. None to load all.
            float32 (bool, default to False): Whether to use float32 or not.
            **kwargs: arguments to be sent to slicing function.

        Returns:
            dict: Subset of data.
        """

        # Make sure fields is not a single element
        if isinstance(fields, str):
            fields = [fields]

        # Make sure partType is not a single element
        if isinstance(partType, str):
            partType = [partType]

        for j, d in enumerate(self._datasets):
            if func == "box":
                r = d.box(kwargs["boundary"], 
                    partType, fields, mdi, float32)
            elif func == "sphere":
                r = d.sphere(kwargs["center"], kwargs["radius"], 
                    partType, fields, mdi, float32)
            else:
                raise ValueError("func must be either \"box\" or \"sphere\"!")

            if j == 0:
                result = r
            else:
                # Loop over particle types
                for p in partType:
                    ptNum = partTypeNum(p)
                    gName = "PartType%d"%(ptNum)

                    # Loop over each requested field for this particle type
                    for i, field in enumerate(fields):
                        # read data local to the current file
                        if mdi is None or mdi[i] is None:
                            result[gName][field] = (np.r_[result[gName][field], 
                                r[gName][field][:]])
                        else:
                            result[gName][field] = (np.r_[result[gName][field], 
                                r[gName][field][:,mdi[i]]])

        return result
    
    def box(self, boundary, partType, fields, mdi=None, float32=False):
        """
        Load a sub-box of data.

        Args:
            boundary (numpy.ndarray of scalar): Boundary of the box, with 
                shape of (3, 2).
            partType (str or list of str): Particle types to be loaded.
            fields (str or list of str): Particle fields to be loaded.
            mdi (None or list of int, default to None): sub-indeces to be 
                loaded. None to load all.
            float32 (bool, default to False): Whether to use float32 or not.

        Returns:
            dict: Sub-box of data.
        """
        return self._combine("box", partType, fields, mdi, float32, 
            boundary=boundary)

    def sphere(self, center, radius, partType, fields, mdi=None, 
        float32=False):
        """
        Load a sub-sphere of data.

        Args:
            center (numpy.ndarray of scalar): Center of the sphere, with 
                shape of (3,).
            radius (scalar): Radius of the sphere.
            partType (str or list of str): Particle types to be loaded.
            fields (str or list of str): Particle fields to be loaded.
            mdi (None or list of int, default to None): sub-indeces to be 
                loaded. None to load all.
            float32 (bool, default to False): Whether to use float32 or not.

        Returns:
            dict: Sub-sphere of data.
        """
        return self._combine("sphere", partType, fields, mdi, float32, 
            center=center, radius=radius)
        
class SingleDataset(object):
    """SingleDataset class stores a chunck of snapshot."""

    def __init__(self, fn, partType, depth=8, index_path=None):
        """
        Args:
            fn (str): File name to be loaded.
            partType (str or list of str): Particle types to be loaded.
            depth (int, default to 8): Depth of Mesh. For example, depth = 8
                corresponds to the Mesh dimension of (2^8, 2^8, 2^8).
            index_path (str): Path to store the index files. None to store 
                with the data.
        """

        super(SingleDataset, self).__init__()

        self._fn = fn

        # Make sure partType is not a single element
        if isinstance(partType, str):
            partType = [partType]
        self._partType = partType

        self._depth = depth

        pt_idx = 0
        for p in partType:
            pt_idx += 2**partTypeNum(p)

        self._index_path = index_path
        suffix = ".idx_d%02d_pt%02d.h5"%(depth, pt_idx)
        if index_path:
            self._index_fn = index_path + fn[fn.rfind("/"):] + suffix
        else:
            self._index_fn = fn + suffix

        self._index = None
        with h5py.File(fn, 'r') as f:
            self._box_size = f['Header'].attrs['BoxSize']
            self._boundary = np.array([[0., 0., 0.],
                [self._box_size, self._box_size, self._box_size]])

        # Set the int type for Mesh
        if depth <= 10:
            # By setting dtype to int32, the maximum level is 10
            self._int_tree = np.int32
        elif depth <=20:
            # By setting dtype to int64, the maximum level is 20
            self._int_tree = np.int64
        else:
            raise ValueError("The depth of Mesh must be no more than 20!")

    @property
    def fn(self):
        """str: File name to be loaded."""
        return self._fn

    @property
    def partType(self):
        """str or list of str: Particle types to be loaded."""
        return self._partType

    @property
    def box_size(self):
        """scalar: Box size of the simulation."""
        return self._box_size

    @property
    def index(self):
        """dict: Newly generated or cached index of the Dataset. """

        if self._index:
            return self._index

        # Generate index
        self._index = {}

        # Load index if index file exists
        if os.path.isfile(self._index_fn):
            with h5py.File(self._index_fn,'r') as f:
                for p in self._partType:
                    ptNum = partTypeNum(p)
                    gName = "PartType%d"%(ptNum)
                    self._index[gName] = {}
                    self._index[gName]["count"] = f[gName].attrs["count"]

                    if not self._index[gName]["count"]:
                        continue

                    self._index[gName]["index"] = f[gName]["index"][:]
                    self._index[gName]["mark"] = f[gName]["mark"][:]

            return self._index

        # Compute and save index if index file does not exist
        data = loadFile(self._fn, self._partType, "Coordinates")
        with h5py.File(self._index_fn, "w") as f:
            # Loop over particle types
            for p in self._partType:
                ptNum = partTypeNum(p)
                gName = "PartType%d"%(ptNum)
                self._index[gName] = {}
                grp = f.create_group(gName)

                length = data[gName]["count"]
                self._index[gName]["count"] = length
                grp.attrs["count"] =  length

                pos = data[gName]["Coordinates"]
                ot = Mesh(pos, length, 0, self._boundary, self._depth)

                self._index[gName]["index"], self._index[gName]["mark"] = ot.build()
                grp.create_dataset("index", 
                    data=self._index[gName]["index"], dtype=np.int64)
                grp.create_dataset("mark", 
                    data=self._index[gName]["mark"], dtype=np.int64)

        return self._index

    def box(self, boundary, partType, fields, mdi=None, float32=True, 
        method="outer"):
        """
        Slicing method to load a sub-box of data.

        Note: The current version only support loading the outer or inner 
            box of the sub-box. Loading the exact sub-box is not supported.

        Args:
            boundary (numpy.ndarray of scalar): Boundary of the box, with 
                shape of (3, 2).
            partType (str or list of str): Particle types to be loaded.
            fields (str or list of str): Particle fields to be loaded.
            mdi (None or list of int, default to None): sub-indeces to be 
                loaded. None to load all.
            float32 (bool, default to False): Whether to use float32 or not.
            method (str, default to "outer"): How to load the box, must be 
                "outer" or "exact" or "inner".

        Returns:
            dict: Sub-box of data.
        """

        # Make sure fields is not a single element
        if isinstance(fields, str):
            fields = [fields]

        # Make sure partType is not a single element
        if isinstance(partType, str):
            partType = [partType]

        self.index # pre-indexing

        boundary_normalized = (
            2**self._depth * (boundary - self._boundary[0]) / 
            (self._boundary[1] - self._boundary[0]))

        if method in ["outer", "exact"]:
            lower = np.floor(boundary_normalized[0]).astype(self._int_tree)
            upper = np.ceil(boundary_normalized[1]).astype(self._int_tree)

        if method == "inner":
            lower = np.ceil(boundary_normalized[0]).astype(self._int_tree)
            upper = np.floor(boundary_normalized[1]).astype(self._int_tree)

        targets = []
        tt0 = 0
        # Use for loop here assuming the box is small
        for p in partType:
            ptNum = partTypeNum(p)
            gName = "PartType%d"%(ptNum)

            t0 = time.time()
            target = _slicing(lower, upper, self._index[gName]["mark"], 
                self._index[gName]["index"], self._depth, self._int_tree)
            tt0 += time.time() - t0

            targets.append(target)

        print("time: %.3fs"%tt0)
        return loadFile(self._fn, partType, fields, mdi, float32, targets)


    def sphere(self, center, radius, partType, fields, mdi=None, 
        method="outer"):
        """
        Slicing method to load a sub-sphere of data.

        Note: This function is not supported in the current version

        Args:
            center (numpy.ndarray of scalar): Center of the sphere, with 
                shape of (3,).
            radius (scalar): Radius of the sphere.
            partType (str or list of str): Particle types to be loaded.
            fields (str or list of str): Particle fields to be loaded.
            mdi (None or list of int, default to None): sub-indeces to be 
                loaded. None to load all.
            float32 (bool, default to False): Whether to use float32 or not.
            method (str, default to "outer"): How to load the box, must be 
                "outer" or "exact" or "inner".
        """
        pass


# Speeding up slicing with numba.jit
from numba import jit, typed, types

@jit(nopython=True)
def _slicing(lower, upper, mark, index, depth, int_tree):
    """
    Slice the index file according to lower/upper boundaries.
    """
    target = typed.List.empty_list(types.int64)
    shifter = np.array([4**depth,2**depth,1], dtype=int_tree)
    for i in range(lower[0], upper[0]):
        for j in range(lower[1], upper[1]):
            idx_3d_lower = np.array([i, j, lower[2]], dtype=int_tree)
            idx_1d_lower = np.sum(idx_3d_lower * shifter)
            idx_3d_upper = np.array([i, j, upper[2]], dtype=int_tree)
            idx_1d_upper = np.sum(idx_3d_upper * shifter)

            start = mark[idx_1d_lower]
            end = mark[idx_1d_upper]
            target.extend(index[start:end])

    return target


# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)

"""
test_il_util module tests APIs in il_util module.
"""

import h5py
import pytest

from mesh_illustris.il_util import *

@pytest.mark.skip(reason="Test function is not ready.")
def test_loadFile():
    pass

@pytest.mark.parametrize("partType, is_error, expected", [
    (0, False, 0),
    ("0", False, 0),
    ("PartType0", False, 0),
    ("stars", False, 4),
    ("Stars", False, 4),
    # Errors
    ("ether", True, "Unknown particle type name.")])
def test_partTypeNum(partType, is_error, expected):
    if is_error:
        with pytest.raises(ValueError, match=expected):
            partTypeNum(partType)
    else:
        assert partTypeNum(partType) == expected

@pytest.mark.parametrize("basePath, snapNum, chunkNum, expected", [
    ("/output", 0, 0, "/output/snapdir_000/snap_000.0.hdf5"),
    ("/output", 135, 99, "/output/snapdir_135/snap_135.99.hdf5")])
def test_snapPath(basePath, snapNum, chunkNum, expected):
    assert snapPath(basePath, snapNum, chunkNum) == expected

@pytest.mark.skip(reason="Test function is not ready.")
def test_getNumPart():
    pass


# -*- coding: utf-8 -*-
# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)
# 
# Part of this file is based on the illustris_python project:
# Copyright (c) 2017 illustris & illustris_python developers
# License: FreeBSD (see licenses/license_illustris_python)

"""
loader module defines a convenient load() function to load snapshots in 
Illustris or IllustrisTNG.
"""

import numpy as np
import h5py

from .core import Dataset, SingleDataset
from .il_util import partTypeNum, snapPath, getNumPart

__all__ = ["load"]

def load(basePath, snapNum, partType, depth=8, index_path=None):
    """
    Function to load snapshots in Illustris or IllustrisTNG.

    Args:
        basePath (str): Base path of the simulation data. This path usually 
            ends with "output".
        snapNum (int): Number of the snapshot.
        partType (str or list of str): Particle types to be loaded.
        depth (int, default to 8): Depth of mesh. For example, depth = 8
            corresponds to the mesh dimension of (2^8, 2^8, 2^8).
        index_path (str): Path to store the index files. None to store 
            with the data.

    Returns:
        `Dataset`: Structured data.
    """

    # Determine number of chunks
    with h5py.File(snapPath(basePath, snapNum), "r") as f:
        n_chunk = f["Header"].attrs["NumFilesPerSnapshot"]

    d = []
    # Loop over chunks
    for i in range(n_chunk):
        fn = snapPath(basePath, snapNum, i)
        d.append(SingleDataset(fn, partType, depth, index_path))

    return Dataset(d, n_chunk)


# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)

"""
run module provides a testing API.
"""

import os

class Runner(object):
    """Runner class provides a testing API."""

    def __init__(self, base_path):
        """
        Args:
            base_path (str): Base path to perform tests. 
        """

        super(Runner, self).__init__()
        self._base_path = os.path.abspath(base_path)
        
    def generate_runner(self):
        """
        Generate the _test_runner function fot self._base_path

        Returns:
            func: A function that calls pytest.main
        """

        def _test_runner(args=[], plugins=[]):
            if not len(args) or not os.path.exists(args[-1]):
                args.insert(len(args), self._base_path)

            # Do not import pytest when it's not needed!
            try:
                import pytest
            except ImportError:
                msg = ("pytest is needed for test(). Install it with "
                    "`pip install pytest`")
                raise ImportError(msg)
            
            return pytest.main(args=args, plugins=plugins)

        return _test_runner


# -*- coding: utf-8 -*-
# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)
# 
# Part of this file is based on the illustris_python project:
# Copyright (c) 2017 illustris & illustris_python developers
# License: FreeBSD (see licenses/license_illustris_python)

"""
il_util module defines some commonly used functions for Illustris.
"""

import numpy as np
import h5py

__all__ = ["loadFile", "partTypeNum", "snapPath", "getNumPart"]

def loadFile(fn, partType, fields=None, mdi=None, float32=True, index=None):
    """
    Load a subset of particles/cells in one chunk file. 
    This function applies numpy.memmap to minimize memory usage.

    Args:
        fn (str): File name to be loaded.
        partType (str or list of str): Particle types to be loaded.
        fields (str or list of str): Particle fields to be loaded.
        mdi (None or list of int, default to None): sub-indeces to be 
            loaded. None to load all.
        float32 (bool, default to False): Whether to use float32 or not.
        index (list of list of int): List of Fancy indices for slicing.

    Returns:
        dict: Entire or subset of data, depending on whether index == None.
    """

    # Make sure fields is not a single element
    if isinstance(fields, str):
        fields = [fields]

    # Make sure partType is not a single element
    if isinstance(partType, str):
        partType = [partType]
    
    result = {}
    with h5py.File(fn, "r") as f:
        for p in partType:
            ptNum = partTypeNum(p)
            gName = "PartType%d"%(ptNum)
            result[gName] = {}

            numType = f["Header"].attrs["NumPart_ThisFile"][ptNum]
            result[gName]["count"] = numType
            if not numType:
                continue

            # Loop over each requested field for this particle type
            for i, field in enumerate(fields):
                # Allocate within return dict
                dtype = f[gName][field].dtype
                shape = f[gName][field].shape
                if dtype == np.float64 and float32: dtype = np.float32
                result[gName][field] = np.zeros(shape, dtype=dtype)

                # read data local to the current file
                ds = f[gName][field]
                offset = ds.id.get_offset()
                dtype = ds.dtype
                shape = ds.shape
                to_load = np.memmap(fn, mode="r", shape=shape, offset=offset, 
                    dtype=dtype)
                if index:
                    if mdi is None or mdi[i] is None:
                        result[gName][field] = to_load[index[i]]
                    else:
                        result[gName][field] = to_load[index[i],mdi[i]]
                else:
                    if mdi is None or mdi[i] is None:
                        result[gName][field] = to_load[:]
                    else:
                        result[gName][field] = to_load[:,mdi[i]]

    return result

def partTypeNum(partType):
    """
    Map common names to numeric particle types.

    Args:
        partType (str): Common names of particle type.

    Returns:
        int: Numeric particle type.
    """

    if str(partType)[-1].isdigit():
        return int(str(partType)[-1])
        
    if str(partType).lower() in ["gas","cells"]:
        return 0
    if str(partType).lower() in ["dm","darkmatter"]:
        return 1
    if str(partType).lower() in ["tracer","tracers","tracermc","trmc"]:
        return 3
    if str(partType).lower() in ["star","stars","stellar"]:
        return 4 # only those with GFM_StellarFormationTime>0
    if str(partType).lower() in ["wind"]:
        return 4 # only those with GFM_StellarFormationTime<0
    if str(partType).lower() in ["bh","bhs","blackhole","blackholes"]:
        return 5
    
    raise ValueError("Unknown particle type name.")

def snapPath(basePath, snapNum, chunkNum=0):
    """
    Return path to a chunk file of snapshot.

    Args:
        basePath (str): Base path of the simulation data. This path usually 
            ends with "output".
        snapNum (int): Number of the snapshot.
        chunkNum (int, default to 0): Number of the chunk.

    Returns:
        str: Path to a chunk file.
    """

    snapPath = basePath + "/snapdir_%03d/"%(snapNum)
    filePath = snapPath + "snap_%03d.%d.hdf5"%(snapNum, chunkNum)

    return filePath

def getNumPart(header):
    """
    Calculate number of particles of all types given a snapshot header.

    Args:
        header (h5py.Group): header of the snapshot.

    Returns:
        numpy.ndarray of int: Number of particles of each type.
    """

    nTypes = 6

    nPart = np.zeros(nTypes, dtype="i4")
    for j in range(nTypes):
        nPart[j] = (header["NumPart_Total"][j] | 
            (header["NumPart_Total_HighWord"][j] << 32))

    return nPart


# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)

from . import core, il_util, loader, mesh
from .core import *
from .il_util import *
from .loader import *
from .mesh import *

__all__ = core.__all__ + il_util.__all__ + loader.__all__ + mesh.__all__
__version__ = "0.2"
__name__ = "mesh_illustris"
__author__ = ["Bill Chen"]

# Initialize test function
from .tests.run import Runner
test = Runner(__path__[0]).generate_runner()
__all__.append("test")




# -*- coding: utf-8 -*-
# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)

from setuptools import setup, find_packages

long_description = (
    "`mesh_illustris` is a toolkit for analyzing "
    "[Illustris](https://www.illustris-project.org/) "
    "(and also [IllustrisTNG](https://www.tng-project.org/)) "
    "data with mesh. The goal of `mesh_illustris` is to "
    "**quickly** load a subset (e.g., a box or sphere) of "
    "particles/cells with **minimal** amount of memory. "
    "Documentation is now available at "
    "[Read the Docs](https://mesh-illustris.readthedocs.io)!")

setup(
    name = 'mesh_illustris',
    packages = find_packages(),
    version = '0.2',
    url = "https://github.com/EnthalpyBill/mesh_illustris",
    license = "MIT",
    author = "Bill Chen <ybchen@umich.edu>",
    maintainer = "Bill Chen <ybchen@umich.edu>",
    description = "Load Illustris with mesh.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    install_requires = ["numpy>=1.18", "h5py>=2.10", "numba>=0.50"],
    python_requires = ">=3.8",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)


# -*- coding: utf-8 -*-
# Copyright (c) 2021 Bill Chen
# License: MIT (see LICENSE)

"""
mesh module defines the Mesh to tessellate the entire volume.
"""

import numpy as np

__all__ = ["Mesh"]

class Mesh(object):
    """Mesh class tessellates the entire volume."""

    def __init__(self, pos, length, offset, boundary, depth):
        """
        Args:
            pos (numpy.ndarray of scalar): Positions of points, with shape of 
                (length, 3).
            length (int): Number of points to be tessellated.
            offset (int): Starting index of points. Often used when there are
                too many points, which need to be separated to many groups.
            boundary (numpy.ndarray of scalar): Boundary of the box, with 
                shape of (3, 2).
            depth (int, default to 8): Depth of Mesh. For example, depth = 8
                corresponds to the Mesh dimension of (2^8, 2^8, 2^8).
        """

        super(Mesh, self).__init__()

        self._pos = pos
        self._length = length
        self._offset = offset
        self._boundary = boundary
        self._depth = depth

        # Set the int type for Mesh
        if depth <= 10:
            # By setting dtype to int32, the maximum level is 10
            self._int_tree = np.int32
        elif depth <=20:
            # By setting dtype to int64, the maximum level is 20
            self._int_tree = np.int64
        else:
            raise ValueError("The depth of Mesh must be no more than 20!")

    @property
    def boundary(self):
        """numpy.ndarray of scalar: Boundary of the box, with 
            shape of (3, 2)."""
        return self._boundary

    @property
    def depth(self):
        """int, default to 8: Depth of Mesh. For example, depth = 8
                corresponds to the Mesh dimension of (2^8, 2^8, 2^8)"""
        return self._depth

    def build(self):
        """
        Build index for the points according to the Mesh.
        The indexing process produces a "rand" and a "mark" variables, which 
        link the index of each point to its location in the Mesh.

        Returns:
            tuple of numpy.ndarray of int: (rank, mark).
        """

        idx_3d = (2**self._depth * (self._pos - self._boundary[0]) //
            (self._boundary[1] - self._boundary[0])).astype(self._int_tree)

        # Conbine 3D index into 1D 
        idx_1d = np.sum(
            np.left_shift(idx_3d, [2*self._depth,self._depth,0]), axis=1)

        # Sort rank with index
        idx = np.argsort(idx_1d)
        idx_all = np.arange(8**self._depth, dtype=self._int_tree)

        mark = np.zeros(8**self._depth+1, dtype=np.int64)
        mark[1:] = np.searchsorted(idx_1d, idx_all, side="right", sorter=idx)

        rank = np.arange(self._offset, self._offset+self._length, 
            dtype=np.int64)
        rank = rank[idx]

        return rank, mark


    
    